#!/usr/bin/env python
"""
Appendix F. Grid Mappings
---
Each recognized grid mapping is described in one of the sections below. Each
section contains: the valid name that is used with the grid_mapping_name
attribute; a list of the specific attributes that may be used to assign values
to the mapping's parameters; the standard names used to identify the coordinate
variables that contain the mapping's independent variables; and references to
the mapping's definition or other information that may help in using the
mapping. Since the attributes used to set a mapping's parameters may be shared
among several mappings, their definitions are contained in a table in the final
section. The attributes which describe the ellipsoid and prime meridian may be
included, when applicable, with any grid mapping.

The grid_mapping_attr_types* dictionaries relate type and condition information
for each of the grid_mapping attributes as defined in the CF standard. Each key-
value pair is comprised of

    <attribute name> : {type: S|N|D, extra_condition: True|False}

where the 'type' signifies the type that the attribute should be, and the
'extra_condition' boolean indicates whether to test for an additional condition
or series of conditions to determine if the value of the attribute is valid.

The grid_mapping_dict* dictionaries map the names of grid mappings to an array
of length 1, 3, or 4. The array is comprised of tuples:

    [
        tuple of required attributes,
        tuple of something that is unclear, (often nothing)
        tuple of expected standard names,
        tuple of exclusive attributes

    ]

We have used the FGDC "Content Standard for Digital Geospatial Metadata" [FGDC]
as a guide in choosing the values for grid_mapping_name and the attribute names
for the parameters describing map projections.
"""

grid_mapping_attr_types16 = {
    "earth_radius": {"type": "N", "extra_condition": False},
    "false_easting": {"type": "N", "extra_condition": False},
    "false_northing": {"type": "N", "extra_condition": False},
    "grid_mapping_name": {"type": "S", "extra_condition": False},
    "grid_north_pole_latitude": {"type": "N", "extra_condition": False},
    "grid_north_pole_longitude": {"type": "N", "extra_condition": False},
    "inverse_flattening": {"type": "N", "extra_condition": False},
    "latitude_of_projection_origin": {"type": "N", "extra_condition": True},
    "longitude_of_central_meridian": {"type": "N", "extra_condition": True},
    "longitude_of_prime_meridian": {"type": "N", "extra_condition": True},
    "longitude_of_projection_origin": {"type": "N", "extra_condition": True},
    "north_pole_grid_longitude": {"type": "N", "extra_condition": False},
    "perspective_point_height": {"type": "N", "extra_condition": False},
    "scale_factor_at_central_meridian": {"type": "N", "extra_condition": True},
    "scale_factor_at_projection_origin": {"type": "N", "extra_condition": True},
    "semi_major_axis": {"type": "N", "extra_condition": False},
    "semi_minor_axis": {"type": "N", "extra_condition": False},
    "standard_parallel": {"type": "N", "extra_condition": True},
    "straight_vertical_longitude_from_pole": {"type": "N", "extra_condition": True},
}

grid_mapping_attr_types17 = (
    grid_mapping_attr_types16.copy()
)  # need shallow copy; update() returns None

grid_mapping_attr_types17.update(
    {
        "azimuth_of_central_line": {"type": "N", "extra_condition": False},
        "crs_wkt": {"type": "S", "extra_condition": False},
        "geographic_crs_name": {"type": "S", "extra_condition": False},
        "geoid_name": {"type": "S", "extra_condition": False},
        "geopotential_datum_name": {"type": "S", "extra_condition": False},
        "horizontal_datum_name": {"type": "S", "extra_condition": True},
        "prime_meridian_name": {"type": "S", "extra_condition": True},
        "projected_crs_name": {"type": "S", "extra_condition": True},
        "reference_ellipsoid_name": {"type": "S", "extra_condition": False},
        "towgs84": {"type": "N", "extra_condition": True},
    }
)


grid_mapping_dict16 = {
    "albers_conical_equal_area": [
        ("longitude_of_central_meridian", "latitude_of_projection_origin"),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
    "azimuthal_equidistant": [
        ("longitude_of_projection_origin", "latitude_of_projection_origin"),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
    "lambert_cylindrical_equal_area": [
        ("longitude_of_central_meridian"),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
        ("standard_parallel", "scale_factor_at_projection_origin"),
    ],
    "lambert_azimuthal_equal_area": [
        ("longitude_of_projection_origin", "latitude_of_projection_origin"),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
    "lambert_conformal_conic": [
        (
            "standard_parallel",
            "longitude_of_central_meridian",
            "latitude_of_projection_origin",
        ),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
    "latitude_longitude": [(), (), ("longitude", "latitude")],
    "mercator": [
        ("longitude_of_projection_origin"),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
        ("standard_parallel", "scale_factor_at_projection_origin"),
    ],
    "orthographic": [
        ("longitude_of_projection_origin", "latitude_of_projection_origin"),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
    "polar_stereographic": [
        ("straight_vertical_longitude_from_pole", "latitude_of_projection_origin"),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
        ("standard_parallel", "scale_factor_at_projection_origin"),
    ],
    "rotated_latitude_longitude": [
        ("grid_north_pole_latitude", "grid_north_pole_longitude"),
        ("north_pole_grid_longitude"),
        ("grid_latitude", "grid_longitude"),
    ],
    "stereographic": [
        (
            "longitude_of_projection_origin",
            "latitude_of_projection_origin",
            "scale_factor_at_projection_origin",
        ),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
    "transverse_mercator": [
        (
            "scale_factor_at_central_meridian",
            "longitude_of_central_meridian",
            "latitude_of_projection_origin",
        ),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
    "vertical_perspective": [
        (
            "longitude_of_projection_origin",
            "latitude_of_projection_origin",
            "perspective_point_height",
        ),
        ("false_easting", "false_northing"),
        ("projection_x_coordinate", "projection_y_coordinate"),
    ],
}

grid_mapping_dict17 = (
    grid_mapping_dict16.copy()
)  # need shallow copy; update() returns None
grid_mapping_dict17.update(
    {
        "geostationary": [
            (
                "latitude_of_projection_origin",
                "longitude_of_projection_origin",
                "perspective_point_height",
            ),
            ("false_easting", "false_northing"),
            ("projection_x_coordinate", "projection_y_coordinate"),
        ],
        "oblique_mercator": [
            (
                "azimuth",
                "latitude_of_projection_origin",
                "longitude_of_projection_origin",
                "scale_factor_at_projection_origin",
            ),
            ("false_easting", "false_northing"),
            ("projection_x_coordinate", "projection_y_coordinate"),
        ],
        "sinusoidal": [
            ("longitude_of_projection_origin"),
            ("false_easting", "false_northing"),
            ("projection_x_coordinate", "projection_y_coordinate"),
        ],
    }
)

# horizontal datum names from https://github.com/cf-convention/cf-conventions/wiki/csv/horiz_datum.csv
horizontal_datum_names17 = {
    "Hungarian Datum 1909",
    "Taiwan Datum 1967",
    "Taiwan Datum 1997",
    "Iraqi Geospatial Reference System",
    "MGI 1901",
    "MOLDREF99",
    "Reseau Geodesique de la RDC 2005",
    "Serbian Reference Network 1998",
    "Red Geodesica de Canarias 1995",
    "Reseau Geodesique de Mayotte 2004",
    "Cadastre 1997",
    "Reseau Geodesique de Saint Pierre et Miquelon 2006",
    "Autonomous Regions of Portugal 2008",
    "Mexican Datum of 1993",
    "China 2000",
    "Sao Tome",
    "New Beijing",
    "Principe",
    "Reseau de Reference des Antilles Francaises 1991",
    "Tokyo 1892",
    "System Jednotne Trigonometricke Site Katastralni/05",
    "Sri Lanka Datum 1999",
    "System Jednotne Trigonometricke Site Katastralni/05 (Ferro)",
    "Geocentric Datum Brunei Darussalam 2009",
    "Turkish National Reference Frame",
    "Bhutan National Geodetic Datum",
    "Islands Net 2004",
    "International Terrestrial Reference Frame 2008",
    "Posiciones Geodesicas Argentinas 2007",
    "Marco Geodesico Nacional",
    "SIRGAS-Chile",
    "Costa Rica 2005",
    "Sistema Geodesico Nacional de Panama MACARIO SOLIS",
    "Peru96",
    "SIRGAS-ROU98",
    "SIRGAS_ES2007.8",
    "Ocotepeque 1935",
    "Sibun Gorge 1922",
    "Panama-Colon 1911",
    "Reseau Geodesique des Antilles Francaises 2009",
    "Corrego Alegre 1961",
    "South American Datum 1969(96)",
    "Papua New Guinea Geodetic Datum 1994",
    "Not specified (based on Airy 1830 ellipsoid)",
    "Not specified (based on Airy Modified 1849 ellipsoid)",
    "Not specified (based on Australian National Spheroid)",
    "Not specified (based on Bessel 1841 ellipsoid)",
    "Not specified (based on Bessel Modified ellipsoid)",
    "Not specified (based on Bessel Namibia ellipsoid)",
    "Not specified (based on Clarke 1858 ellipsoid)",
    "Not specified (based on Clarke 1866 ellipsoid)",
    "Not specified (based on Clarke 1866 Michigan ellipsoid)",
    "Not specified (based on Clarke 1880 (Benoit) ellipsoid)",
    "Not specified (based on Clarke 1880 (IGN) ellipsoid)",
    "Not specified (based on Clarke 1880 (RGS) ellipsoid)",
    "Not specified (based on Clarke 1880 (Arc) ellipsoid)",
    "Not specified (based on Clarke 1880 (SGA 1922) ellipsoid)",
    "Not specified (based on Everest 1830 (1937 Adjustment) ellipsoid)",
    "Not specified (based on Everest 1830 (1967 Definition) ellipsoid)",
    "Not specified (based on Everest 1830 Modified ellipsoid)",
    "Not specified (based on GRS 1980 ellipsoid)",
    "Not specified (based on Helmert 1906 ellipsoid)",
    "Not specified (based on Indonesian National Spheroid)",
    "Not specified (based on International 1924 ellipsoid)",
    "Not specified (based on Krassowsky 1940 ellipsoid)",
    "Not specified (based on NWL 9D ellipsoid)",
    "Not specified (based on Plessis 1817 ellipsoid)",
    "Not specified (based on Struve 1860 ellipsoid)",
    "Not specified (based on War Office ellipsoid)",
    "Not specified (based on WGS 84 ellipsoid)",
    "Not specified (based on GEM 10C ellipsoid)",
    "Not specified (based on OSU86F ellipsoid)",
    "Not specified (based on OSU91A ellipsoid)",
    "Not specified (based on Clarke 1880 ellipsoid)",
    "Not specified (based on Authalic Sphere)",
    "Not specified (based on GRS 1967 ellipsoid)",
    "Not specified (based on Average Terrestrial System 1977 ellipsoid)",
    "Not specified (based on Everest (1830 Definition) ellipsoid)",
    "Not specified (based on WGS 72 ellipsoid)",
    "Not specified (based on Everest 1830 (1962 Definition) ellipsoid)",
    "Not specified (based on Everest 1830 (1975 Definition) ellipsoid)",
    "Not specified (based on GRS 1980 Authalic Sphere)",
    "Not specified (based on Clarke 1866 Authalic Sphere)",
    "Not specified (based on International 1924 Authalic Sphere)",
    "Not specified (based on Hughes 1980 ellipsoid)",
    "Popular Visualisation Datum",
    "Greek",
    "Greek Geodetic Reference System 1987",
    "Average Terrestrial System 1977",
    "Kartastokoordinaattijarjestelma (1966)",
    "Rikets koordinatsystem 1990",
    "Samboja",
    "Lithuania 1994 (ETRS89)",
    "Tete",
    "Madzansua",
    "Observatario",
    "Moznet (ITRF94)",
    "Indian 1960",
    "Final Datum 1958",
    "Estonia 1992",
    "PDO Survey Datum 1993",
    "Old Hawaiian",
    "St. Lawrence Island",
    "St. Paul Island",
    "St. George Island",
    "Puerto Rico",
    "NAD83 Canadian Spatial Reference System",
    "Israel",
    "Locodjo 1965",
    "Abidjan 1987",
    "Kalianpur 1937",
    "Kalianpur 1962",
    "Kalianpur 1975",
    "Hanoi 1972",
    "Hartebeesthoek94",
    "CH1903",
    "CH1903+",
    "Swiss Terrestrial Reference Frame 1995",
    "NAD83 (High Accuracy Reference Network)",
    "Rassadiran",
    "European Datum 1950(1977)",
    "Dabola 1981",
    "System Jednotne Trigonometricke Site Katastralni",
    "Mount Dillon",
    "Naparima 1955",
    "European Libyan Datum 1979",
    "Chos Malal 1914",
    "Pampa del Castillo",
    "Korean Datum 1985",
    "Yemen National Geodetic Network 1996",
    "South Yemen",
    "Bissau",
    "Korean Datum 1995",
    "New Zealand Geodetic Datum 2000",
    "Accra",
    "American Samoa 1962",
    "Sistema de Referencia Geocentrico para America del Sur 1995",
    "Reseau Geodesique Francais 1993",
    "Posiciones Geodesicas Argentinas",
    "IRENET95",
    "Sierra Leone Colony 1924",
    "Sierra Leone 1968",
    "Australian Antarctic Datum 1998",
    "Pulkovo 1942(83)",
    "Pulkovo 1942(58)",
    "Estonia 1997",
    "Luxembourg 1930",
    "Azores Occidental Islands 1939",
    "Azores Central Islands 1948",
    "Azores Oriental Islands 1940",
    "Madeira 1936",
    "OSNI 1952",
    "Red Geodesica Venezolana",
    "Posiciones Geodesicas Argentinas 1998",
    "Albanian 1987",
    "Douala 1948",
    "Manoca 1962",
    "Qornoq 1927",
    "Scoresbysund 1952",
    "Ammassalik 1958",
    "Garoua",
    "Kousseri",
    "Egypt 1930",
    "Pulkovo 1995",
    "Adindan",
    "Australian Geodetic Datum 1966",
    "Australian Geodetic Datum 1984",
    "Ain el Abd 1970",
    "Afgooye",
    "Agadez",
    "Lisbon 1937",
    "Aratu",
    "Arc 1950",
    "Arc 1960",
    "Batavia",
    "Barbados 1938",
    "Beduaram",
    "Beijing 1954",
    "Reseau National Belge 1950",
    "Bermuda 1957",
    "Bogota 1975",
    "Bukit Rimpah",
    "Camacupa",
    "Campo Inchauspe",
    "Cape",
    "Carthage",
    "Chua",
    "Corrego Alegre 1970-72",
    "Cote d'Ivoire",
    "Deir ez Zor",
    "Douala",
    "Egypt 1907",
    "European Datum 1950",
    "European Datum 1987",
    "Fahud",
    "Gandajika 1970",
    "Garoua",
    "Guyane Francaise",
    "Hu Tzu Shan 1950",
    "Hungarian Datum 1972",
    "Indonesian Datum 1974",
    "Indian 1954",
    "Indian 1975",
    "Jamaica 1875",
    "Jamaica 1969",
    "Kalianpur 1880",
    "Kandawala",
    "Kertau 1968",
    "Kuwait Oil Company",
    "La Canoa",
    "Provisional South American Datum 1956",
    "Lake",
    "Leigon",
    "Liberia 1964",
    "Lome",
    "Luzon 1911",
    "Hito XVIII 1963",
    "Herat North",
    "Mahe 1971",
    "Makassar",
    "European Terrestrial Reference System 1989",
    "Malongo 1987",
    "Manoca",
    "Merchich",
    "Massawa",
    "Minna",
    "Mhast",
    "Monte Mario",
    "M'poraloko",
    "North American Datum 1927",
    "NAD27 Michigan",
    "North American Datum 1983",
    "Nahrwan 1967",
    "Naparima 1972",
    "New Zealand Geodetic Datum 1949",
    "NGO 1948",
    "Datum 73",
    "Nouvelle Triangulation Francaise",
    "NSWC 9Z-2",
    "OSGB 1936",
    "OSGB 1970 (SN)",
    "OS (SN) 1980",
    "Padang 1884",
    "Palestine 1923",
    "Congo 1960 Pointe Noire",
    "Geocentric Datum of Australia 1994",
    "Pulkovo 1942",
    "Qatar 1974",
    "Qatar 1948",
    "Qornoq",
    "Loma Quintana",
    "Amersfoort",
    "South American Datum 1969",
    "Sapper Hill 1943",
    "Schwarzeck",
    "Segora",
    "Serindung",
    "Sudan",
    "Tananarive 1925",
    "Timbalai 1948",
    "TM65",
    "Geodetic Datum of 1965",
    "Tokyo",
    "Trinidad 1903",
    "Trucial Coast 1948",
    "Voirol 1875",
    "Bern 1938",
    "Nord Sahara 1959",
    "Stockholm 1938",
    "Yacare",
    "Yoff",
    "Zanderij",
    "Militar-Geographische Institut",
    "Reseau National Belge 1972",
    "Deutsches Hauptdreiecksnetz",
    "Conakry 1905",
    "Dealul Piscului 1930",
    "Dealul Piscului 1970",
    "National Geodetic Network",
    "Kuwait Utility",
    "World Geodetic System 1972",
    "WGS 72 Transit Broadcast Ephemeris",
    "World Geodetic System 1984",
    "Anguilla 1957",
    "Antigua 1943",
    "Dominica 1945",
    "Grenada 1953",
    "Montserrat 1958",
    "St. Kitts 1955",
    "St. Lucia 1955",
    "St. Vincent 1945",
    "North American Datum 1927 (1976)",
    "North American Datum 1927 (CGQ77)",
    "Xian 1980",
    "Hong Kong 1980",
    "Japanese Geodetic Datum 2000",
    "Gunung Segara",
    "Qatar National Datum 1995",
    "Porto Santo 1936",
    "Selvagem Grande",
    "South American Datum 1969",
    "SWEREF99",
    "Point 58",
    "Fort Marigot",
    "Guadeloupe 1948",
    "Centre Spatial Guyanais 1967",
    "Reseau Geodesique Francais Guyane 1995",
    "Martinique 1938",
    "Reunion 1947",
    "Reseau Geodesique de la Reunion 1992",
    "Tahiti 52",
    "Tahaa 54",
    "IGN72 Nuku Hiva",
    "K0 1949",
    "Combani 1950",
    "IGN56 Lifou",
    "IGN72 Grande Terre",
    "ST87 Ouvea",
    "Petrels 1972",
    "Pointe Geologie Perroud 1950",
    "Saint Pierre et Miquelon 1950",
    "MOP78",
    "Reseau de Reference des Antilles Francaises 1991",
    "IGN53 Mare",
    "ST84 Ile des Pins",
    "ST71 Belep",
    "NEA74 Noumea",
    "Reseau Geodesique Nouvelle Caledonie 1991",
    "Grand Comoros",
    "International Terrestrial Reference Frame 1988",
    "International Terrestrial Reference Frame 1989",
    "International Terrestrial Reference Frame 1990",
    "International Terrestrial Reference Frame 1991",
    "International Terrestrial Reference Frame 1992",
    "International Terrestrial Reference Frame 1993",
    "International Terrestrial Reference Frame 1994",
    "International Terrestrial Reference Frame 1996",
    "International Terrestrial Reference Frame 1997",
    "International Terrestrial Reference Frame 2000",
    "Reykjavik 1900",
    "Hjorsey 1955",
    "Islands Net 1993",
    "Helle 1954",
    "Latvia 1992",
    "Porto Santo 1995",
    "Azores Oriental Islands 1995",
    "Azores Central Islands 1995",
    "Lisbon 1890",
    "Iraq-Kuwait Boundary Datum 1992",
    "European Datum 1979",
    "Istituto Geografico Militaire 1995",
    "Voirol 1879",
    "Chatham Islands Datum 1971",
    "Chatham Islands Datum 1979",
    "Sistema de Referencia Geocentrico para las AmericaS 2000",
    "Guam 1963",
    "Vientiane 1982",
    "Lao 1993",
    "Lao National Datum 1997",
    "Jouik 1961",
    "Nouakchott 1965",
    "Mauritania 1999",
    "Gulshan 303",
    "Philippine Reference System 1992",
    "Gan 1970",
    "Gandajika",
    "Marco Geocentrico Nacional de Referencia",
    "Reseau Geodesique de la Polynesie Francaise",
    "Fatu Iva 72",
    "IGN63 Hiva Oa",
    "Tahiti 79",
    "Moorea 87",
    "Maupiti 83",
    "Nakhl-e Ghanem",
    "Posiciones Geodesicas Argentinas 1994",
    "Katanga 1955",
    "Kasai 1953",
    "IGC 1962 Arc of the 6th Parallel South",
    "IGN 1962 Kerguelen",
    "Le Pouce 1934",
    "IGN Astro 1960",
    "Institut Geographique du Congo Belge 1955",
    "Mauritania 1999",
    "Missao Hidrografico Angola y Sao Tome 1951",
    "Mhast (onshore)",
    "Mhast (offshore)",
    "Egypt Gulf of Suez S-650 TL",
    "Tern Island 1961",
    "Cocos Islands 1965",
    "Iwo Jima 1945",
    "St. Helena 1971",
    "Marcus Island 1952",
    "Ascension Island 1958",
    "Ayabelle Lighthouse",
    "Bellevue",
    "Camp Area Astro",
    "Phoenix Islands 1966",
    "Cape Canaveral",
    "Solomon 1968",
    "Easter Island 1967",
    "Fiji Geodetic Datum 1986",
    "Fiji 1956",
    "South Georgia 1968",
    "Grand Cayman 1959",
    "Diego Garcia 1969",
    "Johnston Island 1961",
    "Little Cayman 1961",
    "Midway 1961",
    "Pico de las Nieves 1984",
    "Pitcairn 1967",
    "Santo 1965",
    "Viti Levu 1916",
    "Marshall Islands 1960",
    "Wake Island 1952",
    "Tristan 1968",
    "Kusaie 1951",
    "Deception Island",
    "Geocentric datum of Korea",
    "Hong Kong 1963",
    "Hong Kong 1963(67)",
    "Parametrop Zemp 1990",
    "Faroe Datum 1954",
    "Geodetic Datum of Malaysia 2000",
    "Karbala 1979",
    "Nahrwan 1934",
    "Rauenberg Datum/83",
    "Potsdam Datum/83",
    "Greenland 1996",
    "Vanua Levu 1915",
    "Reseau Geodesique de Nouvelle Caledonie 91-93",
    "ST87 Ouvea",
    "Kertau (RSO)",
    "Viti Levu 1912",
    "fk89",
    "Libyan Geodetic Datum 2006",
    "Datum Geodesi Nasional 1995",
    "Vietnam 2000",
    "SVY21",
    "Jamaica 2001",
    "NAD83 (National Spatial Reference System 2007)",
    "World Geodetic System 1966",
    "Croatian Terrestrial Reference System",
    "Bermuda 2000",
    "Pitcairn 2006",
    "Ross Sea Region Geodetic Datum 2000",
    "Slovenia Geodetic Datum 1996",
    "CH1903 (Bern)",
    "Bogota 1975 (Bogota)",
    "Lisbon 1937 (Lisbon)",
    "Makassar (Jakarta)",
    "Militar-Geographische Institut (Ferro)",
    "Monte Mario (Rome)",
    "Nouvelle Triangulation Francaise (Paris)",
    "Padang 1884 (Jakarta)",
    "Reseau National Belge 1950 (Brussels)",
    "Tananarive 1925 (Paris)",
    "Voirol 1875 (Paris)",
    "Batavia (Jakarta)",
    "Stockholm 1938 (Stockholm)",
    "Greek (Athens)",
    "Carthage (Paris)",
    "NGO 1948 (Oslo)",
    "System Jednotne Trigonometricke Site Katastralni (Ferro)",
    "Nord Sahara 1959 (Paris)",
    "Gunung Segara (Jakarta)",
    "Voirol 1879 (Paris)",
    "International Terrestrial Reference Frame 2005",
    "Ancienne Triangulation Francaise (Paris)",
    "Nord de Guerre (Paris)",
    "Madrid 1870 (Madrid)",
    "Lisbon 1890 (Lisbon)",
}

# prime meridian_names from https://github.com/cf-convention/cf-conventions/wiki/csv/prime_meridian.csv
prime_meridian_names17 = {
    "Athens",
    "Bern",
    "Bogota",
    "Brussels",
    "Ferro",
    "Greenwich",
    "Jakarta",
    "Lisbon",
    "Madrid",
    "Oslo",
    "Paris",
    "Paris RGS",
    "Rome",
    "Stockholm",
}


# ellipsoid names from https://github.com/cf-convention/cf-conventions/wiki/csv/ellipsoid.csv
ellipsoid_names17 = {
    "Airy 1830",
    "Airy Modified 1849",
    "Australian National Spheroid",
    "Average Terrestrial System 1977",
    "Bessel 1841",
    "Bessel Modified",
    "Bessel Namibia",
    "Bessel Namibia (GLM)",
    "CGCS2000",
    "Clarke 1858",
    "Clarke 1866",
    "Clarke 1866 Authalic Sphere",
    "Clarke 1866 Michigan",
    "Clarke 1880",
    "Clarke 1880 (Arc)",
    "Clarke 1880 (Benoit)",
    "Clarke 1880 (IGN)",
    "Clarke 1880 (RGS)",
    "Clarke 1880 (SGA 1922)",
    "Clarke 1880 (international foot)",
    "Danish 1876",
    "Everest (1830 Definition)",
    "Everest 1830 (1937 Adjustment)",
    "Everest 1830 (1962 Definition)",
    "Everest 1830 (1967 Definition)",
    "Everest 1830 (1975 Definition)",
    "Everest 1830 (RSO 1969)",
    "Everest 1830 Modified",
    "GEM 10C",
    "GRS 1967",
    "GRS 1967 Modified",
    "GRS 1980",
    "GRS 1980 Authalic Sphere",
    "Helmert 1906",
    "Hough 1960",
    "Hughes 1980",
    "IAG 1975",
    "Indonesian National Spheroid",
    "International 1924",
    "International 1924 Authalic Sphere",
    "Krassowsky 1940",
    "NWL 9D",
    "OSU86F",
    "OSU91A",
    "PZ-90",
    "Plessis 1817",
    "Popular Visualisation Sphere",
    "Sphere",
    "Struve 1860",
    "WGS 72",
    "WGS 84",
    "War Office",
}



import tempfile

from netCDF4 import Dataset


class MockNetCDF(Dataset):
    """
    Wrapper object around NetCDF Dataset to write data only to memory.
    """

    def __init__(self):
        # taken from test/tst_diskless.py NetCDF library
        # even though we aren't persisting data to disk, the constructor
        # requires a filename not currently in use by another Dataset object..
        tmp_filename = tempfile.NamedTemporaryFile(suffix=".nc", delete=True).name
        super(MockNetCDF, self).__init__(
            tmp_filename, "w", diskless=True, persist=False
        )


class MockTimeSeries(MockNetCDF):
    """
    Mock time series with time dimension and time, lon, lat, and depth
    variables defined
    """

    def __init__(self, default_fill_value=None):
        super(MockTimeSeries, self).__init__()
        self.createDimension("time", 500)
        for v in ("time", "lon", "lat", "depth"):
            self.createVariable(v, "d", ("time",), fill_value=default_fill_value)

        # give some applicable units
        self.variables["time"].units = "seconds since 2019-04-11T00:00:00"
        self.variables["time"].axis = "T"
        self.variables["lat"].units = "degree_north"
        self.variables["lat"].axis = "Y"
        self.variables["lon"].units = "degree_east"
        self.variables["lon"].axis = "X"
        self.variables["depth"].units = "meters"
        self.variables["depth"].axis = "Z"
        self.variables["depth"].positive = "down"


class MockVariable(object):
    """
    For mocking a dataset variable. Constructor optionally takes a NetCDF
    variable, the NetCDF attributes of which will be copied over to this
    object.
    """

    def __init__(self, copy_var=None):
        if copy_var is not None:
            self.name = copy_var.name
            self.dtype = copy_var.dtype
            self.dimensions = copy_var.dimensions
            self.ndim = copy_var.ndim
            for att in copy_var.ncattrs():
                setattr(self, att, getattr(copy_var, att))

    def ncattrs(self):
        return [
            att
            for att in vars(self)
            if att not in {"ndim", "name", "dtype", "dimensions"}
        ]


class MockRaggedArrayRepr(MockNetCDF):
    """
    Class to construct a fake NetCDF dataset using the ragged
    array structure. User must specify whether the structure should
    be contiguous ragged array or incomplete ragged array.
    The user must also specify the featureType of the data set;
    compound featureTypes are acceptable, and all are case-insensitive.

    Users should note that if a compound featureType is given,
    only the valid-CF representation is available. This means
    profiles are organized by contiguous ragged array and an index
    variable is present to denote which "station" the profile belongs
    to.

    Data sets which follow the contiguous ragged array structure
    have a variable called the "count variable" denoting how many
    elements are in a particular instance. This variable must have the
    instance dimension as its dimension. The variable will have an
    attribute "sample_dimension" denoting which dimension is the
    sample dimension.

    Data sets which follow the indexed ragged array structure have
    a variable called the "index variable" denoting which instance
    a particular element belongs to. This variable will have an
    attribute "instance_dimension" denoting the sample dimension.
    The variable's dimension is the sample dimension.

    The netCDF4.Dataset interface available to the user will contain
    a very simple representation of the data model, yielding no
    extraneous dimensions/variables.

    Parameters
    ----------
    feature_type: str (
        profile, timeseries, trajectory,
        timeseriesprofile, trajectoryprofile)

    structure: str (contiguous|indexed)
    """

    def __init__(self, feature_type: str, structure="contiguous"):
        super(MockRaggedArrayRepr, self).__init__()

        if structure.lower() not in ("contiguous", "indexed"):
            raise ValueError("Must initialize MockRaggedArray as contiguous or indexed")

        if feature_type.lower() not in {
            "point",
            "profile",
            "timeseries",
            "trajectory",
            "timeseriesprofile",
            "trajectoryprofile",
        }:
            raise ValueError("Must initialize MockRaggedArray with valid featureType")

        is_compound = False
        if feature_type.lower() in {"timeseriesprofile", "trajectoryprofile"}:
            is_compound = True

        # the data will have 10 instances of whatever and 100
        # total elements
        self.createDimension("INSTANCE_DIMENSION", 10)
        self.createDimension("SAMPLE_DIMENSION", 100)

        # create a variable as the cf_role variable; if a compound
        # featureType is given, multiple variables are created
        if is_compound:
            # compound, need another dimension as well; this will
            # become the "station dimension"
            self.createDimension("STATION_DIMENSION", 5)

            # one variable for "timeseries" or "trajectory" which
            # has the station dimension and cf_role
            _var_name = feature_type.lower().split("profile")[0]
            self.createVariable(
                "{}_id_variable".format(_var_name),
                str,
                ("STATION_DIMENSION",),
                fill_value=None,
            )

            # set the cf_role
            self.variables["{}_id_variable".format(_var_name)].setncattr(
                "cf_role", "{}_id".format(_var_name)
            )

            # there will be one for the profile
            self.createVariable(
                "profile_id_variable", str, ("INSTANCE_DIMENSION",), fill_value=None
            )
            self.variables["profile_id_variable"].setncattr("cf_role", "profile_id")

            # will need a station index variable
            self.createVariable(
                "station_index_variable", int, ("INSTANCE_DIMENSION",), fill_value=None
            )

            self.variables["station_index_variable"].setncattr(
                "instance_dimension", "STATION_DIMENSION"
            )

            # also need counter variable, as compound featureTypes
            # are represented by having a contiguous repr for the
            # profiles and the indexed repr for the timeseries/trajectory
            # organization
            self.createVariable(
                "counter_var",
                "i",  # integer type
                ("INSTANCE_DIMENSION",),
                fill_value=None,
            )

            self.variables["counter_var"].setncattr(
                "sample_dimension", "SAMPLE_DIMENSION"
            )

        else:  # just a single featureType
            self.createVariable(
                "{}_id_variable".format(feature_type),
                str,
                ("INSTANCE_DIMENSION",),
                fill_value=None,
            )

            self.variables["{}_id_variable".format(feature_type)].setncattr(
                "cf_role", "{}_id".format(feature_type)
            )

            if structure == "contiguous":

                # create count variable
                self.createVariable(
                    "counter_var",
                    "i",  # integer type
                    ("INSTANCE_DIMENSION",),
                    fill_value=None,
                )

                self.variables["counter_var"].setncattr(
                    "sample_dimension", "SAMPLE_DIMENSION"
                )

            else:

                # create index variable
                self.createVariable(
                    "index_var",
                    "i",  # integer type
                    ("SAMPLE_DIMENSION",),
                    fill_value=None,
                )
                self.variables["index_var"].setncattr(
                    "instance_dimension", "INSTANCE_DIMENSION"
                )



from contextlib import contextmanager
from functools import lru_cache
from tempfile import NamedTemporaryFile
from typing import BinaryIO, Generator

from netCDF4 import Dataset


try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"


class MemoizedDataset(Dataset):
    """
    A NetCDF dataset which has its get_variables_by_attributes call memoized in
    order to speed up repeated calls to the function.  This should only really
    be used against netCDF Datasets opened in 'r' mode, as the attributes should
    not change upon reading the files.
    """

    @lru_cache(128)
    def get_variables_by_attributes(self, **kwargs):
        return super(MemoizedDataset, self).get_variables_by_attributes(**kwargs)


@contextmanager
def tempnc(data: BinaryIO) -> Generator[str, None, None]:
    """
    Create a temporary in-memory NetCDF file using a NamedTemporaryFile.
    Close the file automatically after scope is exited.

    Type aliasing and tempfile creation credit to @ocefpaf
    https://github.com/ioos/compliance-checker/pull/799#discussion_r411420587

    Parameters
    ----------
    data (bytes): raw bytes to store in the NamedTemporaryFile

    Returns
    -------
    context-managed generator
    """
    tmp = None
    try:
        tmp = NamedTemporaryFile(suffix=".nc", prefix="compliance-checker_")
        tmp.write(data)
        tmp.flush()
        yield tmp.name
    finally:
        if tmp is not None:
            tmp.close()



"""
Compliance Checker suite runner
"""

import codecs
import inspect
import itertools
import os
import re
import subprocess
import sys
import textwrap
import warnings

from collections import defaultdict
from datetime import datetime, timezone
from distutils.version import StrictVersion
from operator import itemgetter
from urllib.parse import urlparse

import requests

from lxml import etree as ET
from netCDF4 import Dataset
from owslib.sos import SensorObservationService
from owslib.swe.sensor.sml import SensorML
from pkg_resources import working_set

from compliance_checker import MemoizedDataset, __version__, tempnc
from compliance_checker.base import BaseCheck, GenericFile, Result, fix_return_value
from compliance_checker.cf.cf import CFBaseCheck
from compliance_checker.protocols import cdl, erddap, netcdf, opendap


# Ensure output is encoded as Unicode when checker output is redirected or piped
if sys.stdout.encoding is None:
    sys.stdout = codecs.getwriter("utf8")(sys.stdout)
if sys.stderr.encoding is None:
    sys.stderr = codecs.getwriter("utf8")(sys.stderr)


def extract_docstring_summary(docstring):
    """
    Returns a dedented docstring without parameter information
    :param docstring: A docstring
    :type docstring: str
    :returns: str
    """
    # return a dedented, then indented two spaces docstring with leading and
    # trailing whitespace removed.
    return re.sub(
        r"^(?=.)",
        "  ",
        textwrap.dedent(
            re.split(r"\n\s*:\w", docstring, flags=re.MULTILINE)[0]
        ).strip(),
        flags=re.MULTILINE,
    )


class CheckSuite(object):
    checkers = (
        {}
    )  # Base dict of checker names to BaseCheck derived types, override this in your CheckSuite implementation
    templates_root = "compliance_checker"  # modify to load alternative Jinja2 templates

    def __init__(self, options=None):
        self.col_width = 40
        self.options = options or {}

    @classmethod
    def _get_generator_plugins(cls):
        """
        Return a list of classes from external plugins that are used to
        generate checker classes
        """

        if not hasattr(cls, "suite_generators"):
            gens = working_set.iter_entry_points("compliance_checker.generators")
            cls.suite_generators = [x.resolve() for x in gens]

        return cls.suite_generators

    def _print_suites(self, verbose=0):
        """
        Prints out available check suites.  If the verbose argument is True,
        includes the internal module version number of the check and also displays
        "latest" meta-versions.
        :param check_suite: Check suite object
        :param verbose: Integer indicating whether to print verbose output
        :type verbose: int
        """
        for checker in sorted(self.checkers.keys()):
            version = getattr(self.checkers[checker], "_cc_checker_version", "???")
            if verbose > 0:
                print(" - {} (v{})".format(checker, version))
            elif ":" in checker and not checker.endswith(
                ":latest"
            ):  # Skip the "latest" output
                print(" - {}".format(checker))

    def _print_checker(self, checker_obj):
        """
        Prints each available check and a description with an abridged
        docstring for a given checker object
        :param checker_obj: Checker object on which to operate
        :type checker_obj: subclass of compliance_checker.base.BaseChecker
        """

        check_functions = self._get_checks(checker_obj, defaultdict(lambda: None))
        for c, _ in check_functions:
            print("- {}".format(c.__name__))
            if c.__doc__ is not None:
                u_doc = c.__doc__
                print("\n{}\n".format(extract_docstring_summary(u_doc)))

    @classmethod
    def add_plugin_args(cls, parser):
        """
        Add command line arguments for external plugins that generate checker
        classes
        """
        for gen in cls._get_generator_plugins():
            gen.add_arguments(parser)

    @classmethod
    def load_generated_checkers(cls, args):
        """
        Load checker classes from generator plugins
        """

        for gen in cls._get_generator_plugins():
            checkers = gen.get_checkers(args)
            cls.checkers.update(checkers)

    @classmethod
    def load_all_available_checkers(cls):
        """
        Helper method to retrieve all sub checker classes derived from various
        base classes.
        """
        cls._load_checkers(working_set.iter_entry_points("compliance_checker.suites"))

    @classmethod
    def _load_checkers(cls, checkers):
        """
        Loads up checkers in an iterable into the class checkers dict
        :param checkers: An iterable containing the checker objects
        """

        for c in checkers:
            try:
                check_obj = c.resolve()
                if hasattr(check_obj, "_cc_spec") and hasattr(
                    check_obj, "_cc_spec_version"
                ):
                    check_version_str = ":".join(
                        (check_obj._cc_spec, check_obj._cc_spec_version)
                    )
                    cls.checkers[check_version_str] = check_obj
                # TODO: remove this once all checkers move over to the new
                #       _cc_spec, _cc_spec_version
                else:
                    # if _cc_spec and _cc_spec_version attributes aren't
                    # present, fall back to using name attribute
                    checker_name = getattr(check_obj, "name", None) or getattr(
                        check_obj, "_cc_spec", None
                    )
                    warnings.warn(
                        "Checker for {} should implement both "
                        '"_cc_spec" and "_cc_spec_version" '
                        'attributes. "name" attribute is deprecated. '
                        "Assuming checker is latest version.",
                        DeprecationWarning,
                    )
                    # append "unknown" to version string since no versioning
                    # info was provided
                    cls.checkers["{}:unknown".format(checker_name)] = check_obj

            except Exception as e:
                print("Could not load", c, ":", e, file=sys.stderr)
        # find the latest version of versioned checkers and set that as the
        # default checker for compliance checker if no version is specified
        ver_checkers = sorted([c.split(":", 1) for c in cls.checkers if ":" in c])
        for spec, versions in itertools.groupby(ver_checkers, itemgetter(0)):
            version_nums = [v[-1] for v in versions]
            try:
                latest_version = str(max(StrictVersion(v) for v in version_nums))
            # if the version can't be parsed as a StrictVersion, parse
            # according to character collation
            except ValueError:
                latest_version = max(version_nums)
            cls.checkers[spec] = cls.checkers[spec + ":latest"] = cls.checkers[
                ":".join((spec, latest_version))
            ]

    def _get_checks(self, checkclass, skip_checks):
        """
        Helper method to retrieve check methods from a Checker class.  Excludes
        any checks in `skip_checks`.

        The name of the methods in the Checker class should start with "check_"
        for this method to find them.
        """
        meths = inspect.getmembers(checkclass, inspect.isroutine)
        # return all check methods not among the skipped checks
        returned_checks = []
        for fn_name, fn_obj in meths:
            if fn_name.startswith("check_") and skip_checks[fn_name] != BaseCheck.HIGH:
                returned_checks.append((fn_obj, skip_checks[fn_name]))

        return returned_checks

    def _run_check(self, check_method, ds, max_level):
        """
        Runs a check and appends a result to the values list.
        @param bound method check_method: a given check method
        @param netCDF4 dataset ds
        @param int max_level: check level
        @return list: list of Result objects
        """
        val = check_method(ds)
        if hasattr(val, "__iter__"):
            # Handle OrderedDict when we need to modify results in a superclass
            # i.e. some checks in CF 1.7 which extend CF 1.6 behaviors
            if isinstance(val, dict):
                val_iter = val.values()
            else:
                val_iter = val
            check_val = []
            for v in val_iter:
                res = fix_return_value(
                    v,
                    check_method.__func__.__name__,
                    check_method,
                    check_method.__self__,
                )
                if max_level is None or res.weight > max_level:
                    check_val.append(res)

            return check_val
        else:
            check_val = fix_return_value(
                val, check_method.__func__.__name__, check_method, check_method.__self__
            )
            if max_level is None or check_val.weight > max_level:
                return [check_val]
            else:
                return []

    def _get_check_versioned_name(self, check_name):
        """
        The compliance checker allows the user to specify a
        check without a version number but we want the report
        to specify the version number.

        Returns the check name with the version number it checked
        """
        if ":" not in check_name or ":latest" in check_name:
            check_name = ":".join(
                (check_name.split(":")[0], self.checkers[check_name]._cc_spec_version)
            )
        return check_name

    def _get_check_url(self, check_name):
        """
        Return the check's reference URL if it exists. If not, return empty str.
        @param check_name str: name of the check being run returned by
                               _get_check_versioned_name()
        """
        return getattr(self.checkers[check_name], "_cc_url", "")

    def _get_valid_checkers(self, ds, checker_names):
        """
        Returns a filtered list of 2-tuples: (name, valid checker) based on the ds object's type and
        the user selected names.
        """

        assert len(self.checkers) > 0, "No checkers could be found."

        if len(checker_names) == 0:
            checker_names = list(self.checkers.keys())

        args = [
            (name, self.checkers[name])
            for name in checker_names
            if name in self.checkers
        ]
        valid = []

        all_checked = set(a[1] for a in args)  # only class types
        checker_queue = set(args)
        while len(checker_queue):
            name, a = checker_queue.pop()
            # is the current dataset type in the supported filetypes
            # for the checker class?
            if type(ds) in a().supported_ds:
                valid.append((name, a))

            # add subclasses of SOS checks
            if "ioos_sos" in name:
                for subc in a.__subclasses__():
                    if subc not in all_checked:
                        all_checked.add(subc)
                        checker_queue.add((name, subc))

        return valid

    @classmethod
    def _process_skip_checks(cls, skip_checks):
        """
        Processes an iterable of skip_checks with strings and returns a dict
        with <check_name>: <max_skip_level> pairs
        """

        check_dict = defaultdict(lambda: None)
        # A is for "all", "M" is for medium, "L" is for low
        check_lookup = {"A": BaseCheck.HIGH, "M": BaseCheck.MEDIUM, "L": BaseCheck.LOW}

        for skip_check_spec in skip_checks:
            split_check_spec = skip_check_spec.split(":")
            check_name = split_check_spec[0]
            if len(split_check_spec) < 2:
                check_max_level = BaseCheck.HIGH
            else:
                try:
                    check_max_level = check_lookup[split_check_spec[1]]
                except KeyError:
                    warnings.warn(
                        "Skip specifier '{}' on check '{}' not found,"
                        " defaulting to skip entire check".format(
                            split_check_spec[1], check_name
                        )
                    )
                    check_max_level = BaseCheck.HIGH

            check_dict[check_name] = check_max_level

        return check_dict

    def run(self, ds, skip_checks, *checker_names):
        """
        Runs this CheckSuite on the dataset with all the passed Checker instances.

        Returns a dictionary mapping checker names to a 2-tuple of their grouped scores and errors/exceptions while running checks.
        """

        ret_val = {}
        checkers = self._get_valid_checkers(ds, checker_names)

        if skip_checks is not None:
            skip_check_dict = CheckSuite._process_skip_checks(skip_checks)
        else:
            skip_check_dict = defaultdict(lambda: None)

        if len(checkers) == 0:
            print(
                "No valid checkers found for tests '{}'".format(",".join(checker_names))
            )

        for checker_name, checker_class in checkers:
            # TODO: maybe this a little more reliable than depending on
            #       a string to determine the type of the checker -- perhaps
            #       use some kind of checker object with checker type and
            #       version baked in
            checker_type_name = checker_name.split(":")[0]
            checker_opts = self.options.get(checker_type_name, set())

            # instantiate a Checker object
            try:
                checker = checker_class(options=checker_opts)
            # hacky fix for no options in constructor
            except TypeError:
                checker = checker_class()
            # TODO? : Why is setup(ds) called at all instead of just moving the
            #         checker setup into the constructor?
            # setup method to prep
            checker.setup(ds)

            checks = self._get_checks(checker, skip_check_dict)
            vals = []
            errs = {}  # check method name -> (exc, traceback)

            for c, max_level in checks:
                try:
                    vals.extend(self._run_check(c, ds, max_level))
                except Exception as e:
                    errs[c.__func__.__name__] = (e, sys.exc_info()[2])

            # score the results we got back
            groups = self.scores(vals)

            # invoke finalizer explicitly
            del checker

            ret_val[checker_name] = groups, errs

        return ret_val

    @classmethod
    def passtree(cls, groups, limit):
        for r in groups:
            if r.children:
                x = cls.passtree(r.children, limit)
                if r.weight >= limit and x is False:
                    return False

            if r.weight >= limit and r.value[0] != r.value[1]:
                return False

        return True

    def build_structure(self, check_name, groups, source_name, limit=1):
        """
        Compiles the checks, results and scores into an aggregate structure which looks like:

            {
              "scored_points": 396,
              "low_count": 0,
              "possible_points": 400,
              "testname": "gliderdac",
              "medium_count": 2,
              "source_name": ".//rutgers/ru01-20140120T1444/ru01-20140120T1649.nc",
              "high_count": 0,
              "all_priorities" : [...],
              "high_priorities": [...],
              "medium_priorities" : [...],
              "low_priorities" : [...]
            }

        @param check_name  The test which was run
        @param groups      List of results from compliance checker
        @param source_name Source of the dataset, used for title
        """
        aggregates = {}

        aggregates["scored_points"] = 0
        aggregates["possible_points"] = 0
        high_priorities = []
        medium_priorities = []
        low_priorities = []
        all_priorities = []

        aggregates["high_count"] = 0
        aggregates["medium_count"] = 0
        aggregates["low_count"] = 0

        def named_function(result):
            for child in result.children:
                all_priorities.append(child)
                named_function(child)

        # For each result, bin them into the appropriate category, put them all
        # into the all_priorities category and add up the point values
        for res in groups:
            if res.weight < limit:
                continue
            # If the result has 0 possible points, then it was not valid for
            # this dataset and contains no meaningful information
            if res.value[1] == 0:
                continue
            aggregates["scored_points"] += res.value[0]
            aggregates["possible_points"] += res.value[1]
            if res.weight == 3:
                high_priorities.append(res)
                if res.value[0] < res.value[1]:
                    aggregates["high_count"] += 1
            elif res.weight == 2:
                medium_priorities.append(res)
                if res.value[0] < res.value[1]:
                    aggregates["medium_count"] += 1
            else:
                low_priorities.append(res)
                if res.value[0] < res.value[1]:
                    aggregates["low_count"] += 1
            all_priorities.append(res)
            # Some results have children
            # We don't render children inline with the top three tables, but we
            # do total the points and display the messages
            named_function(res)

        aggregates["high_priorities"] = high_priorities
        aggregates["medium_priorities"] = medium_priorities
        aggregates["low_priorities"] = low_priorities
        aggregates["all_priorities"] = all_priorities
        aggregates["testname"] = self._get_check_versioned_name(check_name)
        aggregates["source_name"] = source_name
        aggregates["scoreheader"] = self.checkers[check_name]._cc_display_headers
        aggregates["cc_spec_version"] = self.checkers[check_name]._cc_spec_version
        aggregates["cc_url"] = self._get_check_url(aggregates["testname"])
        aggregates["report_timestamp"] = datetime.now(timezone.utc).strftime(
            "%Y-%m-%dT%H:%M:%SZ"
        )
        aggregates["cc_version"] = __version__
        return aggregates

    def dict_output(self, check_name, groups, source_name, limit):
        """
        Builds the results into a JSON structure and writes it to the file buffer.

        @param check_name      The test which was run
        @param groups          List of results from compliance checker
        @param output_filename Path to file to save output
        @param source_name     Source of the dataset, used for title
        @param limit           Integer value for limiting output
        """
        aggregates = self.build_structure(check_name, groups, source_name, limit)
        return self.serialize(aggregates)

    def serialize(self, o):
        """
        Returns a safe serializable object that can be serialized into JSON.

        @param o Python object to serialize
        """
        if isinstance(o, (list, tuple)):
            return [self.serialize(i) for i in o]
        if isinstance(o, dict):
            return {k: self.serialize(v) for k, v in o.items()}
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, Result):
            return self.serialize(o.serialize())
        return o

    def checker_html_output(self, check_name, groups, source_name, limit):
        """
        Renders the HTML output for a single test using Jinja2 and returns it
        as a string.

        @param check_name      The test which was run
        @param groups          List of results from compliance checker
        @param source_name     Source of the dataset, used for title
        @param limit           Integer value for limiting output
        """
        from jinja2 import Environment, PackageLoader

        self.j2 = Environment(
            loader=PackageLoader(self.templates_root, "data/templates")
        )
        template = self.j2.get_template("ccheck.html.j2")

        template_vars = self.build_structure(check_name, groups, source_name, limit)
        return template.render(**template_vars)

    def html_output(self, checkers_html):
        """
        Renders the HTML output for multiple tests and returns it as a string.

        @param checkers_html     List of HTML for single tests as returned by
                                 checker_html_output
        """
        # Note: This relies on checker_html_output having been called so that
        # self.j2 is initialised
        template = self.j2.get_template("ccheck_wrapper.html.j2")
        return template.render(checkers=checkers_html)

    def get_points(self, groups, limit):
        score_list = []
        score_only_list = []

        for g in groups:
            if g.weight >= limit:
                score_only_list.append(g.value)

        # checks where all pertinent sections passed
        all_passed = sum(x[0] == x[1] for x in score_only_list)
        out_of = len(score_only_list)

        # sorts lists into high/medium/low order
        score_list.sort(key=lambda x: x.weight, reverse=True)

        return score_list, all_passed, out_of

    def standard_output(self, ds, limit, check_name, groups):
        """
        Generates the Terminal Output for Standard cases

        Returns the dataset needed for the verbose output, as well as the failure flags.
        """

        score_list, points, out_of = self.get_points(groups, limit)
        issue_count = out_of - points

        # Let's add the version number to the check name if it's missing
        check_name = self._get_check_versioned_name(check_name)
        check_url = self._get_check_url(check_name)
        width = 2 * self.col_width

        # NOTE: printing and use of .center()
        # Nested .format() calls should be avoided when possible.
        # As a future enhancement, a string.Template string might work best here
        # but for the time being individual lines are printed and centered with
        # .center()

        print("\n")
        print("-" * width)
        print("IOOS Compliance Checker Report".center(width))
        print("Version {}".format(__version__).center(width))
        print(
            "Report generated {}".format(
                datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
            ).center(width)
        )
        print("{}".format(check_name).center(width))
        print("{}".format(check_url).center(width))
        print("-" * width)
        if issue_count > 0:
            print("Corrective Actions".center(width))
            plural = "" if issue_count == 1 else "s"
            print(
                "{} has {} potential issue{}".format(
                    os.path.basename(ds), issue_count, plural
                )
            )

        return [groups, points, out_of]

    def standard_output_generation(self, groups, limit, points, out_of, check):
        """
        Generates the Terminal Output
        """
        if points < out_of:
            self.reasoning_routine(groups, check, priority_flag=limit)
        else:
            print("All tests passed!")

    def reasoning_routine(self, groups, check, priority_flag=3, _top_level=True):
        """
        print routine performed
        @param list groups: the Result groups
        @param str check: checker name
        @param int priority_flag: indicates the weight of the groups
        @param bool _top_level: indicates the level of the group so as to
                                print out the appropriate header string
        """

        sort_fn = lambda x: x.weight
        groups_sorted = sorted(groups, key=sort_fn, reverse=True)

        # create dict of the groups -> {level: [reasons]}
        result = {
            key: [v for v in valuesiter if v.value[0] != v.value[1]]
            for key, valuesiter in itertools.groupby(groups_sorted, key=sort_fn)
        }
        priorities = self.checkers[check]._cc_display_headers

        def process_table(res, check):
            """Recursively calls reasoning_routine to parse out child reasons
            from the parent reasons.
            @param Result res: Result object
            @param str check: checker name"""

            issue = res.name
            if not res.children:
                reasons = res.msgs
            else:
                child_reasons = self.reasoning_routine(
                    res.children, check, _top_level=False
                )
                # there shouldn't be messages if there are children
                # is this a valid assumption?
                reasons = child_reasons

            return issue, reasons

        # iterate in reverse to the min priority requested;
        # the higher the limit, the more lenient the output
        proc_strs = ""
        for level in range(3, priority_flag - 1, -1):
            level_name = priorities.get(level, level)
            # print headers
            proc_strs = []

            # skip any levels that aren't in the result
            if level not in result:
                continue

            # skip any empty result levels
            if len(result[level]) > 0:
                # only print priority headers at top level, i.e. non-child
                # datasets
                if _top_level:
                    width = 2 * self.col_width
                    print("\n")
                    print("{:^{width}}".format(level_name, width=width))
                    print("-" * width)

                data_issues = [process_table(res, check) for res in result[level]]

                has_printed = False
                for issue, reasons in data_issues:
                    # if this isn't the first printed issue, add a newline
                    # separating this and the previous level
                    if has_printed:
                        print("")
                    # join alphabetized reasons together
                    reason_str = "\n".join(
                        "* {}".format(r) for r in sorted(reasons, key=lambda x: x[0])
                    )
                    proc_str = "{}\n{}".format(issue, reason_str)
                    print(proc_str)
                    proc_strs.append(proc_str)
                    has_printed = True
        return "\n".join(proc_strs)

    def process_doc(self, doc):
        """
        Attempt to parse an xml string conforming to either an SOS or SensorML
        dataset and return the results
        """
        xml_doc = ET.fromstring(doc)
        if xml_doc.tag == "{http://www.opengis.net/sos/1.0}Capabilities":
            ds = SensorObservationService(None, xml=doc)
            # SensorObservationService does not store the etree doc root,
            # so maybe use monkey patching here for now?
            ds._root = xml_doc

        elif xml_doc.tag == "{http://www.opengis.net/sensorML/1.0.1}SensorML":
            ds = SensorML(xml_doc)
        else:
            raise ValueError("Unrecognized XML root element: {}".format(xml_doc.tag))
        return ds

    def generate_dataset(self, cdl_path):
        """
        Use ncgen to generate a netCDF file from a .cdl file
        Returns the path to the generated netcdf file. If ncgen fails, uses
        sys.exit(1) to terminate program so a long stack trace is not reported
        to the user.

        :param str cdl_path: Absolute path to cdl file that is used to generate netCDF file
        """
        if (
            ".cdl" in cdl_path
        ):  # it's possible the filename doesn't have the .cdl extension
            ds_str = cdl_path.replace(".cdl", ".nc")
        else:
            ds_str = cdl_path + ".nc"

        # generate netCDF-4 file
        iostat = subprocess.run(
            ["ncgen", "-k", "nc4", "-o", ds_str, cdl_path], stderr=subprocess.PIPE
        )
        if iostat.returncode != 0:
            # if not successful, create netCDF classic file
            print(
                "netCDF-4 file could not be generated from cdl file with " + "message:"
            )
            print(iostat.stderr.decode())
            print("Trying to create netCDF Classic file instead.")
            iostat = subprocess.run(
                ["ncgen", "-k", "nc3", "-o", ds_str, cdl_path], stderr=subprocess.PIPE
            )
            if iostat.returncode != 0:
                # Exit program if neither a netCDF Classic nor a netCDF-4 file
                # could be created.
                print(
                    "netCDF Classic file could not be generated from cdl file"
                    + "with message:"
                )
                print(iostat.stderr.decode())
                sys.exit(1)
        return ds_str

    def load_dataset(self, ds_str):
        """
        Returns an instantiated instance of either a netCDF file or an SOS
        mapped DS object.

        :param str ds_str: URL of the resource to load
        """
        # If it's a remote URL load it as a remote resource, otherwise treat it
        # as a local resource.
        pr = urlparse(ds_str)
        if pr.netloc:
            return self.load_remote_dataset(ds_str)
        else:
            return self.load_local_dataset(ds_str)

    def check_remote_netcdf(self, ds_str):
        if netcdf.is_remote_netcdf(ds_str):
            response = requests.get(ds_str, allow_redirects=True, timeout=60)
            try:
                return MemoizedDataset(
                    urlparse(response.url).path, memory=response.content
                )
            except OSError as e:
                # handle case when netCDF C libs weren't compiled with
                # in-memory support by using tempfile
                with tempnc(response.content) as _nc:
                    return MemoizedDataset(_nc)

    def load_remote_dataset(self, ds_str):
        """
        Returns a dataset instance for the remote resource, either OPeNDAP or SOS

        :param str ds_str: URL to the remote resource
        """

        url_parsed = urlparse(ds_str)
        # ERDDAP TableDAP request

        nc_remote_result = self.check_remote_netcdf(ds_str)
        if nc_remote_result:
            return nc_remote_result

        # if application/x-netcdf wasn't detected in the Content-Type headers
        # and this is some kind of erddap tabledap form, then try to get the
        # .ncCF file from ERDDAP
        elif "tabledap" in ds_str and not url_parsed.query:
            # modify ds_str to contain the full variable request
            variables_str = opendap.create_DAP_variable_str(ds_str)

            # join to create a URL to an .ncCF resource
            ds_str = "{}.ncCF?{}".format(ds_str, variables_str)

        nc_remote_result = self.check_remote_netcdf(ds_str)
        if nc_remote_result:
            return nc_remote_result

        # if it's just an OPeNDAP endpoint, use that
        elif opendap.is_opendap(ds_str):
            return Dataset(ds_str)

        # Check if the HTTP response is XML, if it is, it's likely SOS so
        # we'll attempt to parse the response as SOS.
        # Some SOS servers don't seem to support HEAD requests.
        # Issue GET instead if we reach here and can't get the response
        response = requests.get(ds_str, allow_redirects=True, timeout=60)
        content_type = response.headers.get("content-type")
        if content_type.split(";")[0] == "text/xml":
            return self.process_doc(response.content)
        else:
            raise ValueError(
                "Unknown service with content-type: {}".format(content_type)
            )

    def load_local_dataset(self, ds_str):
        """
        Returns a dataset instance for the local resource

        :param ds_str: Path to the resource
        """
        if cdl.is_cdl(ds_str):
            ds_str = self.generate_dataset(ds_str)

        if netcdf.is_netcdf(ds_str):
            return MemoizedDataset(ds_str)

        # Assume this is just a Generic File if it exists
        if os.path.isfile(ds_str):
            return GenericFile(ds_str)

        raise ValueError("File is an unknown format")

    def scores(self, raw_scores):
        """
        Transforms raw scores from a single checker into a fully tallied and grouped scoreline.
        """
        grouped = self._group_raw(raw_scores)

        return grouped

    def _group_raw(self, raw_scores, cur=None, level=1):
        """
        Internal recursive method to group raw scores into a cascading score summary.
        Only top level items are tallied for scores.
        @param list raw_scores: list of raw scores (Result objects)
        """

        def trim_groups(r):
            if isinstance(r.name, tuple) or isinstance(r.name, list):
                new_name = r.name[1:]
            else:
                new_name = []

            return Result(r.weight, r.value, new_name, r.msgs)

        # CHECK FOR TERMINAL CONDITION: all raw_scores.name are single length
        # @TODO could have a problem here with scalar name, but probably still works
        terminal = [len(x.name) for x in raw_scores]
        if terminal == [0] * len(raw_scores):
            return []

        def group_func(r):
            """
            Takes a Result object and slices off the first element of its name
            if its's a tuple. Otherwise, does nothing to the name. Returns the
            Result's name and weight in a tuple to be used for sorting in that
            order in a groupby function.
            @param Result r
            @return tuple (str, int)
            """
            if isinstance(r.name, tuple) or isinstance(r.name, list):
                if len(r.name) == 0:
                    retval = ""
                else:
                    retval = r.name[0:1][0]
            else:
                retval = r.name
            return retval, r.weight

        # END INTERNAL FUNCS ##########################################

        # NOTE until this point, *ALL* Results in raw_scores are
        # individual Result objects.

        # sort then group by name, then by priority weighting
        grouped = itertools.groupby(sorted(raw_scores, key=group_func), key=group_func)

        # NOTE: post-grouping, grouped looks something like
        # [(('Global Attributes', 1), <itertools._grouper at 0x7f10982b5390>),
        #  (('Global Attributes', 3), <itertools._grouper at 0x7f10982b5438>),
        #  (('Not a Global Attr', 1), <itertools._grouper at 0x7f10982b5470>)]
        #  (('Some Variable', 2),     <itertools._grouper at 0x7f10982b5400>),

        ret_val = []

        for k, v in grouped:  # iterate through the grouped tuples

            k = k[0]  # slice ("name", weight_val) --> "name"
            v = list(v)  # from itertools._grouper to list

            cv = self._group_raw(list(map(trim_groups, v)), k, level + 1)
            if len(cv):
                # if this node has children, max weight of children + sum of all the scores
                max_weight = max([x.weight for x in cv])
                sum_scores = tuple(map(sum, list(zip(*([x.value for x in cv])))))
                msgs = []

            else:
                max_weight = max([x.weight for x in v])
                sum_scores = tuple(
                    map(sum, list(zip(*([self._translate_value(x.value) for x in v]))))
                )
                msgs = sum([x.msgs for x in v], [])

            ret_val.append(
                Result(
                    name=k, weight=max_weight, value=sum_scores, children=cv, msgs=msgs
                )
            )

        return ret_val

    def _translate_value(self, val):
        """
        Turns shorthand True/False/None checks into full scores (1, 1)/(0, 1)/(0, 0).
        Leaves full scores alone.
        """
        if val is True:
            return (1, 1)
        elif val is False:
            return (0, 1)
        elif val is None:
            return (0, 0)

        return val



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from tempfile import gettempdir

import pytest

from netCDF4 import Dataset

from compliance_checker.cf import util
from compliance_checker.suite import CheckSuite
from compliance_checker.tests import BaseTestCase
from compliance_checker.tests.resources import STATIC_FILES


mult_msgs_diff = "Failed to find the following messages:\n{missing_msgs}\n\n\
        These were the messages captured:\n{found_msgs}\n\
            Please check wording and section names if messages have been altered since this test was written"


class TestCFIntegration(BaseTestCase):
    def setUp(self):
        """
        Initialize the dataset
        """
        self.cs = CheckSuite()
        self.cs.load_all_available_checkers()
        # get current std names table version (it changes)
        self._std_names = util.StandardNameTable()

    # --------------------------------------------------------------------------------
    # Helper Methods
    # --------------------------------------------------------------------------------

    def new_nc_file(self):
        """
        Make a new temporary netCDF file for the scope of the test
        """
        nc_file_path = os.path.join(gettempdir(), "example.nc")
        if os.path.exists(nc_file_path):
            raise IOError("File Exists: %s" % nc_file_path)
        nc = Dataset(nc_file_path, "w")
        self.addCleanup(os.remove, nc_file_path)
        self.addCleanup(nc.close)
        return nc

    def load_dataset(self, nc_dataset):
        """
        Return a loaded NC Dataset for the given path
        """
        if not isinstance(nc_dataset, str):
            raise ValueError("nc_dataset should be a string")

        nc_dataset = Dataset(nc_dataset, "r")
        self.addCleanup(nc_dataset.close)
        return nc_dataset

    def get_results(self, check_results):
        """
        Returns a tuple of the value scored, possible, and a list of messages
        in the result set.
        """
        aggregation = self.cs.build_structure("cf", check_results["cf"][0], "test", 1)
        out_of = 0
        scored = 0
        results = aggregation["all_priorities"]
        for r in results:
            if isinstance(r.value, tuple):
                out_of += r.value[1]
                scored += r.value[0]
            else:
                out_of += 1
                scored += int(r.value)

        # Store the messages
        messages = []
        for r in results:
            messages.extend(r.msgs)

        return scored, out_of, messages

    def test_sldmb_43093_agg(self):
        dataset = self.load_dataset(STATIC_FILES["sldmb_43093_agg"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        expected_messages = [
            u"attribute time:_CoordianteAxisType should begin with a letter and be composed of letters, digits, and underscores",
            u"attribute lat:_CoordianteAxisType should begin with a letter and be composed of letters, digits, and underscores",
            u"attribute lon:_CoordianteAxisType should begin with a letter and be composed of letters, digits, and underscores",
            u"§2.6.2 global attribute history should exist and be a non-empty string",
            u"standard_name temperature is not defined in Standard Name Table v{}".format(
                self._std_names._version
            ),
            u"temperature's auxiliary coordinate specified by the coordinates attribute, precise_lat, is not a variable in this dataset",
            u"temperature's auxiliary coordinate specified by the coordinates attribute, precise_lon, is not a variable in this dataset",
        ]
        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    @pytest.mark.slowtest
    def test_ocos(self):
        dataset = self.load_dataset(STATIC_FILES["ocos"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)

        expected_messages = [
            "AKs's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "AKt's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "AKv's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "latent's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "lwrad's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "salt's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "sensible's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "shflux's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "swrad's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "temp's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "tke's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "u's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_u (A), xi_u (A) (with U: other/unknown; L: unlimited).",
            "ubar's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_u (A), xi_u (A) (with U: other/unknown; L: unlimited).",
            "v's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_v (A), xi_v (A) (with U: other/unknown; L: unlimited).",
            "vbar's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_v (A), xi_v (A) (with U: other/unknown; L: unlimited).",
            "w's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "zeta's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            '§2.6.1 Conventions global attribute does not contain "CF-1.7"',
            "units (None) attribute of 's_w' must be a string compatible with UDUNITS",
            "units (None) attribute of 's_rho' must be a string compatible with UDUNITS",
            "units (None) attribute of 'Cs_w' must be a string compatible with UDUNITS",
            "units (None) attribute of 'user' must be a string compatible with UDUNITS",
            "units (None) attribute of 'Cs_r' must be a string compatible with UDUNITS",
            "CF recommends latitude variable 'lat_rho' to use units degrees_north",
            "CF recommends latitude variable 'lat_u' to use units degrees_north",
            "CF recommends latitude variable 'lat_v' to use units degrees_north",
            "CF recommends latitude variable 'lat_psi' to use units degrees_north",
            "CF recommends longitude variable 'lon_rho' to use units degrees_east",
            "CF recommends longitude variable 'lon_u' to use units degrees_east",
            "CF recommends longitude variable 'lon_v' to use units degrees_east",
            "CF recommends longitude variable 'lon_psi' to use units degrees_east",
            "§4.3.3 The standard_name of `s_rho` must map to the correct computed_standard_name, `['altitude', 'height_above_geopotential_datum', 'height_above_mean_sea_level', 'height_above_reference_ellipsoid']`",
            "§4.3.3 The standard_name of `s_w` must map to the correct computed_standard_name, `['altitude', 'height_above_geopotential_datum', 'height_above_mean_sea_level', 'height_above_reference_ellipsoid']`",
        ]
        assert set(messages).issubset(set(expected_messages))

    def test_l01_met(self):
        dataset = self.load_dataset(STATIC_FILES["l01-met"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        # The variable is supposed to be a status flag but it's mislabled
        expected_messages = [
            "units for variable air_temperature_qc must be convertible to K currently they are 1",
            "units for variable wind_speed_qc must be convertible to m s-1 currently they are 1",
            "standard_name visibility is not defined in Standard Name Table v{}".format(
                self._std_names._version
            ),
            "standard_name modifier data_quality for variable visibility_qc is not a valid modifier according to appendix C",
            "standard_name wind_direction is not defined in Standard Name Table v{}".format(
                self._std_names._version
            ),
            "standard_name modifier data_quality for variable wind_direction_qc is not a valid modifier according to appendix C",
            "standard_name wind_gust is not defined in Standard Name Table v{}".format(
                self._std_names._version
            ),
            "standard_name modifier data_quality for variable wind_gust_qc is not a valid modifier according to appendix C",
            "standard_name modifier data_quality for variable air_temperature_qc is not a valid modifier according to appendix C",
            "standard_name use_wind is not defined in Standard Name Table v{}".format(
                self._std_names._version
            ),
            "standard_name barometric_pressure is not defined in Standard Name Table v{}".format(
                self._std_names._version
            ),
            "standard_name modifier data_quality for variable barometric_pressure_qc is not a valid modifier according to appendix C",
            "standard_name modifier data_quality for variable wind_speed_qc is not a valid modifier according to appendix C",
            "standard_name barometric_pressure is not defined in Standard Name Table v{}".format(
                self._std_names._version
            ),
            "CF recommends latitude variable 'lat' to use units degrees_north",
            "CF recommends longitude variable 'lon' to use units degrees_east",
        ]

        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    def test_usgs_dem_saipan(self):
        dataset = self.load_dataset(STATIC_FILES["usgs_dem_saipan"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        expected_messages = [
            '§2.6.1 Conventions global attribute does not contain "CF-1.7"'
        ]

        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    def test_sp041(self):
        dataset = self.load_dataset(STATIC_FILES["sp041"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of
        assert (u"lat_qc is not a variable in this dataset") in messages

    def test_3mf07(self):
        """Load the 3mf07.nc file and run the CF check suite on it. There should be
        several variable/attribute combos which fail:
          - latitude:valid min
          - latitude:valid_max
          - longitude:valid_min
          - longitude:valid_max
          - references is an empty string
          - comment (global attr) is an empty string
          - z:dimensions are not a proper subset of dims for variable flag, haul
        """

        dataset = self.load_dataset(STATIC_FILES["3mf07"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        expected_messages = [
            u"latitude:valid_min must be a numeric type not a string",
            u"latitude:valid_max must be a numeric type not a string",
            u"longitude:valid_min must be a numeric type not a string",
            u"longitude:valid_max must be a numeric type not a string",
            u"§2.6.2 references global attribute should be a non-empty string",
            u"§2.6.2 comment global attribute should be a non-empty string",
            u"dimensions for auxiliary coordinate variable z (z) are not a subset of dimensions for variable flag (profile)",
            u"dimensions for auxiliary coordinate variable z (z) are not a subset of dimensions for variable haul (profile)",
        ]

        assert scored < out_of
        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    def test_ooi_glider(self):
        dataset = self.load_dataset(STATIC_FILES["ooi_glider"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        expected_messages = [
            u"§2.6.2 comment global attribute should be a non-empty string",
            u"units (None) attribute of 'deployment' must be a string compatible with UDUNITS",
            u"Attribute long_name or/and standard_name is highly recommended for variable deployment",
            u"latitude variable 'latitude' should define standard_name='latitude' or axis='Y'",
            u"longitude variable 'longitude' should define standard_name='longitude' or axis='X'",
        ]

        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    def test_swan(self):
        dataset = self.load_dataset(STATIC_FILES["swan"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        expected_messages = [
            "global attribute _CoordSysBuilder should begin with a letter and be composed of letters, digits, and underscores",
            '§2.6.1 Conventions global attribute does not contain "CF-1.7"',
            "units for variable time_offset must be convertible to s currently they are hours since 2013-02-18T00:00:00Z",
            "units for variable time_run must be convertible to s currently they are hours since 2013-02-18 00:00:00.000 UTC",
            "lon's axis attribute must be T, X, Y, or Z, currently x",
            "lat's axis attribute must be T, X, Y, or Z, currently y",
            "z's axis attribute must be T, X, Y, or Z, currently z",
            "z: vertical coordinates not defining pressure must include a positive attribute that is either 'up' or 'down'",
            "GRID is not a valid CF featureType. It must be one of point, timeseries, trajectory, profile, timeseriesprofile, trajectoryprofile",
        ]

        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    def test_kibesillah(self):
        dataset = self.load_dataset(STATIC_FILES["kibesillah"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of
        # test for global attributes (CF 2.6.2)
        assert (
            u"§2.6.2 global attribute title should exist and be a non-empty string"
        ) in messages

    def test_pr_inundation(self):
        dataset = self.load_dataset(STATIC_FILES["pr_inundation"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        expected_messages = [
            "waterlevel's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "velocity_x's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), Layer (Z), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "velocity_y's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), Layer (Z), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "tau_x's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "tau_y's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "§2.6.2 grid_depth:comment should be a non-empty string",
            "§2.6.2 depth:comment should be a non-empty string",
            "§2.6.2 institution global attribute should be a non-empty string",
            "§2.6.2 comment global attribute should be a non-empty string",
            "units (None) attribute of 'LayerInterf' must be a string compatible with UDUNITS",
            "units (None) attribute of 'time_bounds' must be a string compatible with UDUNITS",
            "units (None) attribute of 'Layer' must be a string compatible with UDUNITS",
            "units for variable area must be convertible to m2 currently they are degrees2",
            "k: vertical coordinates not defining pressure must include a positive attribute that is either 'up' or 'down'",
            "grid_longitude has no coordinate associated with a variable identified as true latitude/longitude; its coordinate variable should also share a subset of grid_longitude's dimensions",
            "grid_latitude has no coordinate associated with a variable identified as true latitude/longitude; its coordinate variable should also share a subset of grid_latitude's dimensions",
            "time_bounds might be a cell boundary variable but there are no variables that define it as a boundary using the `bounds` attribute.",
        ]
        assert set(expected_messages).issubset(messages)

    def test_fvcom(self):
        dataset = self.load_dataset(STATIC_FILES["fvcom"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        for msg in messages:
            if msg.startswith("dimensions for auxiliary coordinate variable siglay"):
                break
        # it's not clear to me what this is supposed to be doing -- this else clause is outside of the if
        else:
            raise AssertionError(
                u'"dimensions for auxiliary coordinate variable siglay (node, siglay) '
                'are not a subset of dimensions for variable u (siglay, nele, time)"'
                " not in messages"
            )
        assert (
            '§2.6.1 Conventions global attribute does not contain "CF-1.7"'
        ) in messages

    def test_ww3(self):
        dataset = self.load_dataset(STATIC_FILES["ww3"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

        expected_messages = [
            u"§2.6.2 global attribute title should exist and be a non-empty string",
            u"§2.6.2 global attribute history should exist and be a non-empty string",
            u"§2.6.1 Conventions field is not present",
            u"Attribute long_name or/and standard_name is highly recommended for variable time",
            u"Attribute long_name or/and standard_name is highly recommended for variable lon",
            u"Attribute long_name or/and standard_name is highly recommended for variable lat",
            u"latitude variable 'lat' should define standard_name='latitude' or axis='Y'",
            u"longitude variable 'lon' should define standard_name='longitude' or axis='X'",
        ]

        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join(
                "\n".join([m for m in expected_messages if m not in messages])
            ),
            found_msgs="\n".join(messages),
        )

    def test_glcfs(self):
        dataset = self.load_dataset(STATIC_FILES["glcfs"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of
        # TODO: referenced/relative time is treated like time units
        assert (
            "units for variable time_offset must be convertible to s currently "
            "they are hours since 2016-01-01T12:00:00Z"
        ) in messages
        assert (
            "standard_name cloud_cover is not defined in Standard Name Table v{}".format(
                self._std_names._version
            )
        ) in messages
        assert (
            u"standard_name dew_point is not defined in Standard Name Table v{}".format(
                self._std_names._version
            )
        ) in messages
        assert (
            u"GRID is not a valid CF featureType. It must be one of point, timeseries, "
            "trajectory, profile, timeseriesprofile, trajectoryprofile"
        ) in messages
        assert (
            u"global attribute _CoordSysBuilder should begin with a letter and "
            "be composed of letters, digits, and underscores"
        ) in messages
        assert u"source should be defined"
        assert (u'units for cl, "fraction" are not recognized by UDUNITS') in messages

    def test_ncei_templates(self):
        """
        Tests some of the NCEI NetCDF templates, which usually should get a
        perfect score.
        """
        dataset = self.load_dataset(STATIC_FILES["NCEI_profile_template_v2_0"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)
        assert scored < out_of

    def test_bad_cf_roles(self):
        """
        Tests the CF checker detects datasets with more than 2 defined cf_role
        variables.
        """
        dataset = self.load_dataset(STATIC_FILES["bad_cf_role"])
        check_results = self.cs.run(dataset, [], "cf")
        scored, out_of, messages = self.get_results(check_results)

        expected_messages = [
            u"§2.6.2 global attribute title should exist and be a non-empty string",
            u"§2.6.2 global attribute history should exist and be a non-empty string",
            u"§2.6.1 Conventions field is not present",
            u"§9.5 The only acceptable values of cf_role for Discrete Geometry CF data sets are timeseries_id, profile_id, and trajectory_id",
        ]

        assert scored < out_of
        assert all([m in messages for m in expected_messages]), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    def test_no_incorrect_errors_index_ragged_array__subset_of_dimensions(
        self,
    ):  # ,wrong_msg):
        """
        From Github issue #845:\n
        CF: incorrect errors for ragged array structure\n
        \n
        Structure needs to be correctly identified\n
        http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#_indexed_ragged_array_representation\n
        https://github.com/ioos/compliance-checker/issues/845
        """
        dataset = self.load_dataset(STATIC_FILES["index_ragged2"])
        check_results = self.cs.run(dataset, [], "cf")
        wrong_msg = "are not a subset of dimensions for variable"
        messages = self.get_results(check_results)[-1]

        assert wrong_msg not in "".join(messages)

    def test_no_incorrect_errors_index_ragged_array__unidentifiable_feature(self):
        """
        From Github issue #845:\n
        CF: incorrect errors for ragged array structure\n
        \n
        Structure needs to be correctly identified\n
        http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#_indexed_ragged_array_representation\n
        https://github.com/ioos/compliance-checker/issues/845
        """
        dataset = self.load_dataset(STATIC_FILES["index_ragged2"])
        check_results = self.cs.run(dataset, [], "cf")
        wrong_msg = "Unidentifiable feature for variable"
        messages = self.get_results(check_results)[-1]

        assert wrong_msg not in "".join(messages)



import io
import itertools
import os
import sys

from collections import defaultdict
from copy import deepcopy
from pkgutil import get_data
from urllib.parse import urljoin

import lxml.html
import requests

from cf_units import Unit
from lxml import etree
from netCDF4 import Dimension, Variable
from pkg_resources import resource_filename


# copied from paegan
# paegan may depend on these later
_possiblet = {
    "time",
    "TIME",
    "Time",
    "t",
    "T",
    "ocean_time",
    "OCEAN_TIME",
    "jd",
    "JD",
    "dn",
    "DN",
    "times",
    "TIMES",
    "Times",
    "mt",
    "MT",
    "dt",
    "DT",
}
_possiblez = {
    "depth",
    "DEPTH",
    "depths",
    "DEPTHS",
    "height",
    "HEIGHT",
    "altitude",
    "ALTITUDE",
    "alt",
    "ALT",
    "Alt",
    "Altitude",
    "h",
    "H",
    "s_rho",
    "S_RHO",
    "s_w",
    "S_W",
    "z",
    "Z",
    "siglay",
    "SIGLAY",
    "siglev",
    "SIGLEV",
    "sigma",
    "SIGMA",
    "vertical",
    "VERTICAL",
    "lev",
    "LEV",
    "level",
    "LEVEL",
}
_possiblex = {
    "x",
    "X",
    "lon",
    "LON",
    "xlon",
    "XLON",
    "lonx",
    "lonx",
    "lon_u",
    "LON_U",
    "lon_v",
    "LON_V",
    "lonc",
    "LONC",
    "Lon",
    "Longitude",
    "longitude",
    "LONGITUDE",
    "lon_rho",
    "LON_RHO",
    "lon_psi",
    "LON_PSI",
}
_possibley = {
    "y",
    "Y",
    "lat",
    "LAT",
    "ylat",
    "YLAT",
    "laty",
    "laty",
    "lat_u",
    "LAT_U",
    "lat_v",
    "LAT_V",
    "latc",
    "LATC",
    "Lat",
    "Latitude",
    "latitude",
    "LATITUDE",
    "lat_rho",
    "LAT_RHO",
    "lat_psi",
    "LAT_PSI",
}

_possibleaxis = _possiblet | _possiblez | _possiblex | _possibley


_possiblexunits = {
    "degrees_east",
    "degree_east",
    "degrees_E",
    "degree_E",
    "degreesE",
    "degreeE",
}

_possibleyunits = {
    "degrees_north",
    "degree_north",
    "degrees_N",
    "degree_N",
    "degreesN",
    "degreeN",
}

_possibletunits = {
    "day",
    "days",
    "d",
    "hour",
    "hours",
    "hr",
    "hrs",
    "h",
    "year",
    "years",
    "minute",
    "minutes",
    "m",
    "min",
    "mins",
    "second",
    "seconds",
    "s",
    "sec",
    "secs",
}

_possibleaxisunits = _possiblexunits | _possibleyunits | _possibletunits


class DotDict(dict):
    """
    Subclass of dict that will recursively look up attributes with dot notation.
    This is primarily for working with JSON-style data in a cleaner way like javascript.
    Note that this will instantiate a number of child DotDicts when you first access attributes;
    do not use in performance-critical parts of your code.
    """

    def __dir__(self):
        return list(self.__dict__.keys()) + list(self.keys())

    def __getattr__(self, key):
        """Make attempts to lookup by nonexistent attributes also attempt key lookups."""
        if key in self:
            return self[key]
        import dis
        import sys

        frame = sys._getframe(1)
        if "\x00%c" % dis.opmap["STORE_ATTR"] in frame.f_code.co_code:
            self[key] = DotDict()
            return self[key]

        raise AttributeError(key)

    def __setattr__(self, key, value):
        if key in dir(dict):
            raise AttributeError("%s conflicts with builtin." % key)
        if isinstance(value, dict):
            self[key] = DotDict(value)
        else:
            self[key] = value

    def copy(self):
        return deepcopy(self)

    def get_safe(self, qual_key, default=None):
        """
        @brief Returns value of qualified key, such as "system.name" or None if not exists.
                If default is given, returns the default. No exception thrown.
        """
        value = get_safe(self, qual_key)
        if value is None:
            value = default
        return value

    @classmethod
    def fromkeys(cls, seq, value=None):
        return DotDict(dict.fromkeys(seq, value))


def get_safe(dict_instance, keypath, default=None):
    """
    Returns a value with in a nested dict structure from a dot separated
    path expression such as "system.server.host" or a list of key entries
    @retval Value if found or None
    """
    try:
        obj = dict_instance
        keylist = keypath if type(keypath) is list else keypath.split(".")
        for key in keylist:
            obj = obj[key]
        return obj
    except Exception:
        return default


class NCGraph(object):
    def __init__(
        self, ds, name, nc_object, self_reference_variables, reference_map=None
    ):

        self.ds = ds
        self.name = name
        self.coords = DotDict()
        self.dims = DotDict()
        self.grid_mapping = DotDict()
        self.obj = nc_object

        self.reference_variables = self_reference_variables
        self.reference_map = reference_map or {}

        self.reference_map[name] = self

        if isinstance(nc_object, Dimension):
            self._type = "dim"

        elif isinstance(nc_object, Variable):
            self._type = "var"

            self.get_references()

        else:
            raise TypeError("unknown type %s" % repr(type(nc_object)))

    def get_references(self):
        for dim in self.obj.dimensions:
            self.dims[dim] = self.get_dimension(dim)

        if hasattr(self.obj, "coordinates"):
            coords = self.obj.coordinates.split(" ")
            for coord in coords:
                self.coords[coord] = self.get_coordinate(coord)

        if hasattr(self.obj, "grid_mapping"):
            gm = self.obj.grid_mapping
            self.grid_mapping[gm] = self.get_grid_mapping(gm)

    def get_dimension(self, dim):
        if dim in self.reference_map:
            return self.reference_map[dim]
        return NCGraph(
            self.ds,
            dim,
            self.ds.dimensions[dim],
            self.reference_variables,
            self.reference_map,
        )

    def get_coordinate(self, coord):
        if coord not in self.ds.variables:
            return
        if coord in self.reference_map:
            if self.name == coord:
                self.reference_variables.add(self.name)
            return self.reference_map[coord]
        return NCGraph(
            self.ds,
            coord,
            self.ds.variables[coord],
            self.reference_variables,
            self.reference_map,
        )

    def get_grid_mapping(self, gm):
        if gm not in self.ds.variables:
            return
        if gm in self.reference_map:
            return self.reference_map[gm]
        return NCGraph(
            self.ds,
            gm,
            self.ds.variables[gm],
            self.reference_variables,
            self.reference_map,
        )

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        return getattr(self.obj, key)


class StandardNameTable(object):
    class NameEntry(object):
        def __init__(self, entrynode):
            self.canonical_units = self._get(entrynode, "canonical_units", True)
            self.grib = self._get(entrynode, "grib")
            self.amip = self._get(entrynode, "amip")
            self.description = self._get(entrynode, "description")

        def _get(self, entrynode, attrname, required=False):
            vals = entrynode.xpath(attrname)
            if len(vals) > 1:
                raise Exception("Multiple attrs (%s) found" % attrname)
            elif required and len(vals) == 0:
                raise Exception("Required attr (%s) not found" % attrname)

            return vals[0].text

    def __init__(self, cached_location=None):
        if cached_location:
            with io.open(cached_location, "r", encoding="utf-8") as fp:
                resource_text = fp.read()
        elif os.environ.get("CF_STANDARD_NAME_TABLE") and os.path.exists(
            os.environ["CF_STANDARD_NAME_TABLE"]
        ):
            with io.open(
                os.environ["CF_STANDARD_NAME_TABLE"], "r", encoding="utf-8"
            ) as fp:
                resource_text = fp.read()
        else:
            resource_text = get_data(
                "compliance_checker", "data/cf-standard-name-table.xml"
            )

        parser = etree.XMLParser(remove_blank_text=True)
        self._root = etree.fromstring(resource_text, parser)

        # generate and save a list of all standard names in file
        self._names = [node.get("id") for node in self._root.iter("entry")]
        self._aliases = [node.get("id") for node in self._root.iter("alias")]
        self._version = self._root.xpath("version_number")[0].text

    def __len__(self):
        return len(self._names) + len(self._aliases)

    def __getitem__(self, key):
        if not (key in self._names or key in self._aliases):
            raise KeyError("%s not found in standard name table" % key)

        if key in self._aliases:
            idx = self._aliases.index(key)
            entryids = self._root.xpath("alias")[idx].xpath("entry_id")

            if len(entryids) != 1:
                raise Exception(
                    "Inconsistency in standard name table, could not lookup alias for %s"
                    % key
                )

            key = entryids[0].text

        if key not in self._names:
            raise KeyError("%s not found in standard name table" % key)

        idx = self._names.index(key)
        entry = self.NameEntry(self._root.xpath("entry")[idx])
        return entry

    def get(self, key, default=None):
        """
        Returns the item for the key or returns the default if it does not exist
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        return key in self._names or key in self._aliases

    def __iter__(self):
        return iter(itertools.chain(self._names, self._aliases))


def download_cf_standard_name_table(version, location=None):
    """
    Downloads the specified CF standard name table version and saves it to file

    :param str version: CF standard name table version number (i.e 34)
    :param str location: Path/filename to write downloaded xml file to
    """

    if (
        location is None
    ):  # This case occurs when updating the packaged version from command line
        location = resource_filename(
            "compliance_checker", "data/cf-standard-name-table.xml"
        )

    if version == "latest":
        tables_tree = lxml.html.parse("http://cfconventions.org/documents.html")
        end_str = "cf-standard-name-table.xml"
        xpath_expr = (
            "//a[substring(@href, string-length(@href) - "
            "string-length('{0}') +1) "
            " = '{0}'][1]".format(end_str)
        )
        latest_vers = tables_tree.xpath(xpath_expr)[0]

        url = urljoin("http://cfconventions.org", latest_vers.attrib["href"])
    else:
        url = "http://cfconventions.org/Data/cf-standard-names/{0}/src/cf-standard-name-table.xml".format(
            version
        )

    r = requests.get(url, allow_redirects=True)
    r.raise_for_status()

    print(
        "Downloading cf-standard-names table version {0} from: {1}".format(
            version, url
        ),
        file=sys.stderr,
    )
    with open(location, "wb") as f:
        f.write(r.content)


def create_cached_data_dir():
    """
    Returns the path to the data directory to download CF standard names.
    Use $XDG_DATA_HOME.
    """
    writable_directory = os.path.join(os.path.expanduser("~"), ".local", "share")
    data_directory = os.path.join(
        os.environ.get("XDG_DATA_HOME", writable_directory), "compliance-checker"
    )
    if not os.path.isdir(data_directory):
        os.makedirs(data_directory)

    return data_directory


def units_known(units):
    try:
        Unit(units)
    except ValueError:
        return False
    return True


def units_convertible(units1, units2, reftimeistime=True):
    """Return True if a Unit representing the string units1 can be converted
    to a Unit representing the string units2, else False."""
    try:
        u1 = Unit(units1)
        u2 = Unit(units2)
    except ValueError:
        return False
    return u1.is_convertible(u2)


def units_temporal(units):
    try:
        u = Unit(units)
    except ValueError:
        return False
    return u.is_time_reference()


def map_axes(dim_vars, reverse_map=False):
    """
    axis name       -> [dimension names]
    dimension name  -> [axis_name], length 0 if reverse_map
    """
    ret_val = defaultdict(list)
    axes = ["X", "Y", "Z", "T"]

    for k, v in dim_vars.items():
        axis = getattr(v, "axis", "")
        if not axis:
            continue

        axis = axis.upper()
        if axis in axes:
            if reverse_map:
                ret_val[k].append(axis)
            else:
                ret_val[axis].append(k)

    return dict(ret_val)


def find_coord_vars(ncds):
    """
    Finds all coordinate variables in a dataset.

    A variable with the same name as a dimension is called a coordinate variable.
    """
    coord_vars = []

    for d in ncds.dimensions:
        if d in ncds.variables and ncds.variables[d].dimensions == (d,):
            coord_vars.append(ncds.variables[d])

    return coord_vars


def is_time_variable(varname, var):
    """
    Identifies if a variable is represents time
    """
    satisfied = varname.lower() == "time"
    satisfied |= getattr(var, "standard_name", "") == "time"
    satisfied |= getattr(var, "axis", "") == "T"
    satisfied |= units_convertible(
        "seconds since 1900-01-01", getattr(var, "units", "")
    )
    return satisfied


def is_vertical_coordinate(var_name, var):
    """
    Determines if a variable is a vertical coordinate variable

    4.3
    A vertical coordinate will be identifiable by: units of pressure; or the presence of the positive attribute with a
    value of up or down (case insensitive).  Optionally, the vertical type may be indicated additionally by providing
    the standard_name attribute with an appropriate value, and/or the axis attribute with the value Z.
    """
    # Known name
    satisfied = var_name.lower() in _possiblez
    satisfied |= getattr(var, "standard_name", "") in _possiblez
    # Is the axis set to Z?
    satisfied |= getattr(var, "axis", "").lower() == "z"
    is_pressure = units_convertible(getattr(var, "units", "1"), "dbar")
    # Pressure defined or positive defined
    satisfied |= is_pressure
    if not is_pressure:
        satisfied |= getattr(var, "positive", "").lower() in ("up", "down")
    return satisfied



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for command line output and parsing

"""
import io
import json
import os
import sys
import tempfile

from argparse import Namespace
from unittest import TestCase

import pytest

from compliance_checker.runner import CheckSuite, ComplianceChecker
from compliance_checker.tests.resources import STATIC_FILES


class TestCLI(TestCase):
    """
    Tests various functions and aspects of the command line tool and runner
    """

    def setUp(self):
        self.fid, self.path = tempfile.mkstemp()
        # why is the class being written to
        CheckSuite.checkers.clear()
        CheckSuite.load_all_available_checkers()

    def tearDown(self):
        if os.path.isfile(self.path):
            os.close(self.fid)
            os.remove(self.path)

    def shortDescription(self):
        return None

    # override __str__ and __repr__ behavior to show a copy-pastable nosetest name for ion tests
    #  ion.module:TestClassName.test_function_name
    def __repr__(self):
        name = self.id()
        name = name.split(".")
        if name[0] not in ["ion", "pyon"]:
            return "%s (%s)" % (name[-1], ".".join(name[:-1]))
        else:
            return "%s ( %s )" % (
                name[-1],
                ".".join(name[:-2]) + ":" + ".".join(name[-2:]),
            )

    __str__ = __repr__

    def test_unicode_acdd_html(self):
        """
        Tests that the checker is capable of producing HTML with unicode characters
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=STATIC_FILES["2dim"],
            verbose=0,
            criteria="strict",
            checker_names=["acdd"],
            output_filename=self.path,
            output_format="html",
        )

        assert os.stat(self.path).st_size > 0

    def test_unicode_cf_html(self):
        """
        Tests that the CF checker can produce HTML output with unicode characters
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=STATIC_FILES["2dim"],
            verbose=0,
            criteria="strict",
            checker_names=["cf"],
            output_filename=self.path,
            output_format="html",
        )

        assert os.stat(self.path).st_size > 0

    def test_single_json_output(self):
        """
        Tests that a suite can produce JSON output to a file
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=STATIC_FILES["conv_bad"],
            verbose=0,
            criteria="strict",
            checker_names=["cf"],
            output_filename=self.path,
            output_format="json",
        )

        assert os.stat(self.path).st_size > 0
        with open(self.path) as f:
            r = json.load(f)
            assert "cf" in r

    def test_list_checks(self):
        """
        Tests listing of both old-style, deprecated checkers using .name
        attributes, and newer ones which use ._cc_spec and _cc_spec_version
        attributes
        """

        # hack: use argparse.Namespace to mock checker object with attributes
        # since SimpleNamespace is Python 3.3+ only
        CheckSuite.checkers.clear()
        # need to mock setuptools entrypoints here in order to load in checkers
        def checker_1():
            return Namespace(name="checker_1")

        def checker_2():
            return Namespace(_cc_spec="checker_2", _cc_spec_version="2.2")

        mock_checkers = [Namespace(resolve=checker_1), Namespace(resolve=checker_2)]
        with pytest.warns(DeprecationWarning):
            CheckSuite._load_checkers(mock_checkers)

        cs = CheckSuite()
        saved = sys.stdout
        try:
            # ugly!  consider refactoring to use py.test capsys
            fake_stdout = io.StringIO()
            sys.stdout = fake_stdout
            cs._print_suites()
            assert fake_stdout.getvalue() == " - checker_1:unknown\n - checker_2:2.2\n"
        finally:
            sys.stdout = saved
            fake_stdout.close()

    def test_multiple_json_output(self):
        """
        Tests that a suite can produce JSON output to a file
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=STATIC_FILES["conv_bad"],
            verbose=0,
            criteria="strict",
            checker_names=["acdd", "cf"],
            output_filename=self.path,
            output_format="json",
        )

        assert os.stat(self.path).st_size > 0
        with open(self.path) as f:
            r = json.load(f)
            assert "cf" in r
            assert "acdd" in r

    def test_multiple_json_output_stdout(self):
        """
        Tests that a suite can produce JSON output to stdout
        """
        saved = sys.stdout
        try:
            fake_stdout = io.StringIO()
            sys.stdout = fake_stdout
            return_value, errors = ComplianceChecker.run_checker(
                ds_loc=STATIC_FILES["conv_bad"],
                verbose=0,
                criteria="strict",
                checker_names=["acdd", "cf"],
                output_filename="-",
                output_format="json",
            )
            r = json.loads(fake_stdout.getvalue().strip())
            assert "acdd" in r
            assert "cf" in r
        finally:
            sys.stdout = saved
            fake_stdout.close()

    def test_single_json_output_stdout(self):
        """
        Tests that a suite can produce JSON output to stdout
        """
        saved = sys.stdout
        try:
            fake_stdout = io.StringIO()
            sys.stdout = fake_stdout
            return_value, errors = ComplianceChecker.run_checker(
                ds_loc=STATIC_FILES["conv_bad"],
                verbose=0,
                criteria="strict",
                checker_names=["cf"],
                output_filename="-",
                output_format="json",
            )
            r = json.loads(fake_stdout.getvalue().strip())
            assert "cf" in r
        finally:
            sys.stdout = saved
            fake_stdout.close()

    def test_text_output(self):
        """
        Tests that the 'text' output can be redirected to file with arguments
        to the command line
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=STATIC_FILES["conv_bad"],
            verbose=0,
            criteria="strict",
            checker_names=["acdd", "cf"],
            output_filename=self.path,
            output_format="text",
        )

        assert os.stat(self.path).st_size > 0

    def test_multi_checker_return_value(self):
        """
        Tests that any failure for multiple checkers results in a failure return
        status
        """
        # CF should pass here
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=STATIC_FILES["ncei_gold_point_1"],
            verbose=0,
            criteria="strict",
            checker_names=["cf:1.6"],
            output_filename=self.path,
            output_format="text",
        )
        self.assertTrue(return_value)

        # CF should pass, but ACDD will have some errors.  Overall return
        # status should be a failure
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=STATIC_FILES["ncei_gold_point_1"],
            verbose=0,
            criteria="strict",
            checker_names=["acdd", "cf"],
            output_filename=self.path,
            output_format="text",
        )
        self.assertFalse(return_value)
